package cz.mendelu.dnaAnalyser.sequence.inport;

import cz.mendelu.dnaAnalyser.batch.BatchService;
import cz.mendelu.dnaAnalyser.sequence.NCBISequenceService;
import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.SequenceRepository;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository;
import cz.mendelu.dnaAnalyser.utils.exception.InternalServerException;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;

import java.io.IOException;
import java.io.InputStream;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ForkJoinPool;
import java.util.regex.Pattern;

/**
 * Created by Jan Kolomazník on 26.6.17.
 * Optimized by Dis. Josef Havránek in March 2019
 */
@Service
public class ImportSequenceService {
    public static final String FASTA_PSEOUDO_LINE="\0";
    private static final String FASTA = "fasta";//todo centralize  data  types
    private static final String PLAIN = "plain";
    private static final Pattern FASTA_SPLITTER = Pattern.compile("[$,;|]");

    private final Pattern pseudoLineCleanup = Pattern.compile(FASTA_PSEOUDO_LINE);
    private final Pattern fastaCleanup = Pattern.compile("^[>;]");

    @Autowired
    private SequenceRepository sequenceRepository;

    @Autowired
    private SequenceDataRepository sequenceDataRepository;
    @Autowired
    private NCBISequenceService ncbiSequenceService;
    @Autowired
    private RawDataProcessor dataProcessor;

    @Autowired
    private BatchService batchService;

    private static final ForkJoinPool pool = SequenceImportPool.importPool;

    public Sequence importPlainFormatFromFile(Sequence sequence, MultipartFile buffer) {
        FASTAPlainPreBatch(sequence);
        try (InputStream data = buffer.getInputStream()) {
            sequenceDataRepository.saveRawBufferFromFile(sequence.getBufferId(), PLAIN, data);
        } catch (IOException e) {
            throw new InternalServerException("can't obtain stream from from imported file", e);
        }
        return commonPlainBatchExecutor(sequence);
    }

    private Sequence commonPlainBatchExecutor(Sequence sequence) {
        batchService.createBatch(sequence.getOwner(), sequence)
                .execute(uuid -> {
                    Sequence s = sequenceRepository
                            .findById(uuid)
                            .orElseThrow(() -> new NotFoundException(Sequence.class, uuid));
                    dataProcessor.batchProcessor(s, PLAIN, RawDataProcessor::transformPlain);
                }).start();
        return sequence;
    }

    public Sequence importPlainFormatFromText(Sequence sequence, byte[] buffer) {
        FASTAPlainPreBatch(sequence);
        sequenceDataRepository.saveRawBuffer(sequence.getBufferId(), PLAIN, buffer);
        return commonPlainBatchExecutor(sequence);
    }

    /**
     * Common operation before batch on Sequences of FASTA and plain format
     */
    private void FASTAPlainPreBatch(Sequence sequence) {
        assert sequence.getOwner() != null : "Sequence must have user!";
        sequence.setBufferId(sequence.getId());
        sequence.setCreated(new Date());
        sequenceRepository.save(sequence);
    }

    public Sequence importFASTAFormatFromFile(Sequence sequence, MultipartFile buffer) {
        FASTAPlainPreBatch(sequence);
        try (InputStream data = buffer.getInputStream()) {
            sequenceDataRepository.saveRawBufferFromFile(sequence.getBufferId(), FASTA, data);
        } catch (IOException e) {
            throw new InternalServerException("can't obtain stream from from imported file", e);
        }
        return commonFASTABatchExecutor(sequence);
    }

    public Sequence importFASTAFormatFromText(Sequence sequence, byte[] buffer) {
        FASTAPlainPreBatch(sequence);
        sequenceDataRepository.saveRawBuffer(sequence.getBufferId(), FASTA, buffer);
        return commonFASTABatchExecutor(sequence);
    }

    private Sequence commonFASTABatchExecutor(Sequence sequence) {
        batchService.createBatch(sequence.getOwner(), sequence)
                .execute(uuid -> {
                    Sequence s = sequenceRepository
                            .findById(uuid)
                            .orElseThrow(() -> new NotFoundException(Sequence.class, uuid));
                    StringBuffer fastaComment = new StringBuffer();
                    dataProcessor.batchProcessor(s, FASTA, RawDataProcessor.transformFASTAToInternal(fastaComment));
                    updateFastaCommentAndName(s, fastaComment);
                }).start();
        return sequence;
    }

    public Sequence importNCBI(Sequence sequence, String ncbiId) {
        assert sequence.getOwner() != null : "Sequence must have user!";

        List<Sequence> sameSequenceList = sequenceRepository.findByNcbi(ncbiId);

        if (sameSequenceList != null && !sameSequenceList.isEmpty()) {
            Sequence sameSequence = sameSequenceList.get(0);
            sequence.setNcbi(ncbiId);
            sequence.setLength(sameSequence.getLength());
            sequence.setBufferId(sameSequence.getBufferId());
            sequence.setFastaComment(sameSequence.getFastaComment());
            sequence.setNucleicCounts(sameSequence.getNucleicCounts());
            sequence = sequenceRepository.save(sequence);
            Logger logger = LoggerFactory.getLogger(ImportSequenceService.class);
            /*System.out.println(" OWNER                      " + sequence.getOwner());
            System.out.println(" ID                      " + sequence.getId().toString());*/
            System.out.println(" IDENTIFIABLE          " + batchService.findBatchFor(sameSequence));

            //System.out.println("BATCH             "  + batchService.findBatchByEntityClassAndEntityId("cz.mendelu.dnaAnalyser.sequence.Sequence","530d53b9-f518-4345-89d2-82afa8ba27eb"));


            return sequence;
        }
        sequence.setCreated(new Date());
        sequence.setBufferId(sequence.getId());
        sequence = sequenceRepository.save(sequence);


        batchService.createBatch(sequence.getOwner(), sequence)
                .execute(uuid -> {
                    Sequence s = sequenceRepository
                            .findById(uuid)
                            .orElseThrow(() -> new NotFoundException(Sequence.class, uuid));

                    StringBuffer fastaComment = new StringBuffer();
                    pool.invoke(
                            dataProcessor.batchProcessorHelper(s,
                                    ncbiSequenceService.downloadFASTA(ncbiId),
                                    RawDataProcessor.transformFASTAToInternal(fastaComment)
                            )
                    );
                    updateFastaCommentAndName(s, fastaComment);
                }).start();
        return sequence;
    }

    private void updateFastaCommentAndName(Sequence sequence, StringBuffer fastaComments) {
        String comments = pseudoLineCleanup.matcher(fastaComments.toString()).replaceAll("\n");
        comments = fastaCleanup.matcher(comments).replaceAll("");
        sequence.setFastaComment(comments);
        if (StringUtils.isBlank(sequence.getName())) {
            sequence.setName(extractNameFromFASTAComment(comments));
        }
    }

    private String extractNameFromFASTAComment(String fastaComment) {
        String[] parts = FASTA_SPLITTER.split(fastaComment);
        return (parts.length > 0)
                ? parts[0].trim()
                : null;
    }

}
