package cz.mendelu.dnaAnalyser.sequence

import cz.mendelu.dnaAnalyser.sequence.data.SequenceData
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository
import spock.lang.Specification

import java.nio.ByteBuffer
import java.nio.file.Path
import java.util.stream.Collectors

class SequenceDataRepositoryTest extends Specification {

    SequenceDataRepository sequenceBufferRepository
    File sequenceBufferDirectory
    UUID sequenceId
    UUID bufferId
    Sequence sequence

    def setup() {
        sequenceBufferRepository = new SequenceDataRepository()

        sequenceBufferDirectory = File.createTempDir('SequenceDataRepositoryTest-', '-createDir')
        sequenceBufferDirectory.deleteOnExit()

        sequenceBufferRepository.setStorageDir(sequenceBufferDirectory.canonicalPath)

        sequenceId = UUID.randomUUID()
        bufferId = UUID.randomUUID()
        sequence = new Sequence(
                id: sequenceId,
                bufferId: bufferId
        )
    }


    def "Create StorageDir if not exists"() {
        when:
        sequenceBufferRepository.setStorageDir(sequenceBufferDirectory.canonicalPath)


        then:
        sequenceBufferDirectory.exists()
        sequenceBufferDirectory.isDirectory()
    }

    def "Save raw buffer"() {
        given:
        def format = 'test'
        def buffer = 'ACTG'.bytes

        when:
        Path file = sequenceBufferRepository.saveRawBuffer(bufferId, format, buffer)

        then:
        sequenceBufferDirectory.listFiles().contains file.toFile()
        file.readBytes() == 'ACTG'.bytes
    }

    def "Load raw buffer"() {
        given:
        def format = 'test'

        and:
        File storedBuffer = new File(sequenceBufferDirectory, "${sequence.bufferId}.test")
        storedBuffer << 'ACTG'
        when:
        def buffer = sequenceBufferRepository.loadRawBufferToStream(sequence, format).collect(Collectors.toList()).get(0)

        then:
        buffer == new String("ACTG")
        buffer.length() == 4
    }

    def "Try override raw file"() {
        given:
        def uuid = UUID.randomUUID()
        def format = 'test'
        def buffer = 'ACTG'.bytes

        and:
        sequenceBufferRepository.saveRawBuffer(uuid, format, buffer)

        when:
        sequenceBufferRepository.saveRawBuffer(uuid, format, buffer)

        then:
        thrown(java.lang.AssertionError)
    }

    def "Save sequence buffer (internal format)"() {
        given:
        def uuid = UUID.randomUUID()
        def buffer = new ArrayList(Arrays.asList(ByteBuffer.wrap('ACTG'.bytes))).iterator()

        when:
        def file = sequenceBufferRepository.save(uuid, buffer).toFile()

        then:
        sequenceBufferDirectory.listFiles().contains file
        file.readBytes() == 'ACTG'.bytes
    }

    def "Load sequence buffer (internal format)"() {
        given:
        File storedBuffer = new File(sequenceBufferDirectory, "${sequence.bufferId}")
        storedBuffer << 'ACTG'

        when:
        SequenceData sequenceBuffer = sequenceBufferRepository.load(sequence)
        then:
        sequenceBuffer.uuid == sequence.id
        sequenceBuffer.length == 4
        sequenceBuffer.toString(0, 4) == 'ACTG'
    }

    def "Delete raw sequence buffer"() {
        given:
        def uuid = UUID.randomUUID()
        def format = 'test'

        and:
        File storedBuffer = new File(sequenceBufferDirectory, "${uuid}.test")
        storedBuffer << 'ACTG'

        when:
        sequenceBufferRepository.deleteRawBuffer(uuid, format)

        then:
        sequenceBufferDirectory.list().contains("${uuid}.test") == false
    }

    def "Delete sequence buffer (internal format)"() {
        given:
        def uuid = UUID.randomUUID()

        and:
        File storedBuffer = new File(sequenceBufferDirectory, "${uuid}")
        storedBuffer << 'ACTG'

        when:
        sequenceBufferRepository.delete(uuid)

        then:
        sequenceBufferDirectory.list().contains("${uuid}") == false
    }
}
